from typing import List
from enum import Enum


class BaseModelName(str, Enum):
    LLAMA_3_1_8B = "llama-3.1-8B"
    LLAMA_3_1_13B = "llama-3.1-13B"
    TINY_LLAMA_1B = "TinyLlama-1B"


class TrainType(str, Enum):
    FINE_TUNE = "fine-tune"
    TRANSFER = "transfer"



import os
import json
import click
from constellaxion.handlers.cloud_job import GCPDeployJob


def get_job(print=False):
    if os.path.exists("job.json"):
        with open("job.json", "r") as f:
            config = json.load(f)
        if print:
            click.echo(click.style(
                "Model Job Config Details:", bold=True, fg="blue"))
            click.echo(json.dumps(config, indent=4))
        return config
    else:
        click.echo(click.style(
            "Error: job.json not found. Run 'constellaxion init' first", fg="red"))
        return None


@click.group()
def job():
    """Manage jobs"""
    pass


@job.command()
def run():
    """Run training job"""
    click.echo(click.style(f"Preparing training job...", fg="blue"))
    config = get_job()
    if config:
        cloud = config['deploy']['provider']
        match cloud:
            case "gcp":
                job = GCPDeployJob()
                job.run(config)


@job.command()
def view():
    """View the status or details of one or more jobs"""
    get_job(print=True)



import os
import argparse
import pandas as pd
from trl import SFTTrainer, DataCollatorForCompletionOnlyLM
from transformers import TrainingArguments, AutoModelForCausalLM, AutoTokenizer
from peft import LoraConfig, TaskType, get_peft_model
from datasets import Dataset
from torch.utils.data import DataLoader
from google.cloud import storage

# Parse cli args
parser = argparse.ArgumentParser()
parser.add_argument("--train-set", type=str, required=True,
                    help="Training set path")
parser.add_argument("--val-set", type=str, required=True,
                    help="Validation set path")
parser.add_argument("--test-set", type=str, required=True,
                    help="Test set path")
parser.add_argument("--bucket-name", type=str, required=True,
                    help="GCS bucket name")
parser.add_argument("--model-path", type=str, required=True,
                    help="Model artefacts output path")
parser.add_argument("--experiments-dir", type=str, required=True,
                    help="Experiments output path")
parser.add_argument("--staging-dir", type=str, required=True,
                    help="Staging path")

args = parser.parse_args()

SEED = 42

MODEL_NAME = 'TinyLlama/TinyLlama-1.1B-intermediate-step-1431k-3T'
PAD_TOKEN = '<pad>'

LOCAL_MODEL_DIR = './model'
GCS_BUCKET_NAME = args.bucket_name
GCS_MODEL_PATH = args.model_path
TRAIN_CSV = f"gs://{GCS_BUCKET_NAME}/{args.train_set}"
VAL_CSV = f"gs://{GCS_BUCKET_NAME}/{args.val_set}"
TEST_CSV = f"gs://{GCS_BUCKET_NAME}/{args.test_set}"
OUTPUT_DIR = args.experiments_dir

# Dataset
train_df = pd.read_csv(TRAIN_CSV)
val_df = pd.read_csv(VAL_CSV)
test_df = pd.read_csv(TEST_CSV)

dataset = {
    "train": Dataset.from_pandas(train_df),
    "val": Dataset.from_pandas(val_df),
    "test": Dataset.from_pandas(test_df)
}

# Tokenizer
tokenizer = AutoTokenizer.from_pretrained(
    MODEL_NAME, add_eos_token=True, use_fast=True)
tokenizer.add_special_tokens({"pad_token": PAD_TOKEN})
tokenizer.padding_side = "right"

# Model
model = AutoModelForCausalLM.from_pretrained(
    MODEL_NAME, device_map="auto", trust_remote_code=True)
model.resize_token_embeddings(len(tokenizer), pad_to_multiple_of=8)

model.pad_token_id = tokenizer.pad_token_id
model.config.pad_token_id = tokenizer.pad_token_id
print(model.config)
print(model)

# LoRA
lora_config = LoraConfig(
    r=128, lora_alpha=128,
    target_modules=[
        "self_attn.q_proj",
        "self_attn.k_proj",
        "self_attn.v_proj",
        "self_attn.o_proj"
    ],
    lora_dropout=0.1,
    bias="none",
    task_type=TaskType.CAUSAL_LM
)

model = get_peft_model(model, lora_config)
model.print_trainable_parameters()

# Prepare data loader
response_template = "\n### Prediction:"
response_template_ids = tokenizer.encode(
    response_template, add_special_tokens=False)

collator = DataCollatorForCompletionOnlyLM(
    response_template_ids, tokenizer=tokenizer)

examples = dataset["train"][0]
encodings = [tokenizer(e) for e in examples]

dataloader = DataLoader(encodings, collate_fn=collator, batch_size=1)
batch = next(iter(dataloader))
print(batch.keys())

# Train Model
train_args = TrainingArguments(
    output_dir=OUTPUT_DIR,
    num_train_epochs=1,
    per_device_train_batch_size=4,
    gradient_accumulation_steps=4,
    optim="adamw_torch",
    evaluation_strategy="steps",
    eval_steps=0.2,
    logging_steps=10,
    learning_rate=1e-4,
    fp16=True,
    save_strategy="epoch",
    max_grad_norm=1.0,
    warmup_ratio=0.1,
    lr_scheduler_type="constant",
    save_safetensors=True,
    seed=SEED
)

trainer = SFTTrainer(
    model=model,
    args=train_args,
    train_dataset=dataset["train"],
    eval_dataset=dataset["val"],
    tokenizer=tokenizer,
    max_seq_length=1024,
    data_collator=collator
)

trainer.train()


def upload_directory_to_gcs(local_path, bucket_name, gcs_path):
    """Upload to GCS"""
    client = storage.Client()
    bucket = client.bucket(bucket_name)

    for root, _, files in os.walk(local_path):
        for file in files:
            local_file_path = os.path.join(root, file)
            relative_path = os.path.relpath(local_file_path, local_path)
            gcs_blob_path = os.path.join(gcs_path, relative_path)

            blob = bucket.blob(gcs_blob_path)
            blob.upload_from_filename(local_file_path)
            print(
                f"Uploaded {local_file_path} to gs://{bucket_name}/{gcs_blob_path}")


def save_model_tokenizer_locally(model, tokenizer, save_dir):
    """Save model and tokenizer locally"""
    os.makedirs(save_dir, exist_ok=True)
    model.save_pretrained(save_dir)
    tokenizer.save_pretrained(save_dir)
    print(f"Model and tokenizer saved locally to {save_dir}")


def save_and_upload_model(model, tokenizer):
    """Saven and upload model"""
    # Save locally
    save_model_tokenizer_locally(model, tokenizer, LOCAL_MODEL_DIR)

    # Upload to GCS
    upload_directory_to_gcs(LOCAL_MODEL_DIR, GCS_BUCKET_NAME, GCS_MODEL_PATH)


save_and_upload_model(trainer.model, tokenizer)



class Set():
    def __init__(self, type: str, path: str):
        self.type = type
        self.path = path


class Dataset():
    def __init__(self, train: str, val: str, test: str):
        self._train = Set('train', train)
        self._val = Set('val', val)
        self._test = Set('test', test)
        self.train = self._get_train()
        self.val = self._get_val()
        self.test = self._get_test()

    def _get_train(self):
        return self._train.path

    def _get_val(self):
        return self._val.path

    def _get_test(self):
        return self._test.path



model_map = {
    "TinyLlama-1B": {
        "dir": "tinyllama_1b",
        "lora": "../../models/tinyllama_1b/gcp/lora.py",
        "infra": {
            "container_uri": "europe-docker.pkg.dev/vertex-ai/training/tf-gpu.2-14.py310:latest",
            "requirements": [
                "trl==0.12.0",
                "transformers",
                "dataset",
                "peft",
                "google-cloud-storage"
            ],
            "machine_type": "g2-standard-4",
            "accelerator_type": "NVIDIA_L4",
            "accelerator_count": 1,
            "replica_count": 1
        }
    }
}



import json
import random
import string
from abc import abstractmethod, ABC
from constellaxion.handlers.model import Model
from constellaxion.handlers.dataset import Dataset
from constellaxion.services.gcp.train_job import run_training_job


class BaseCloudJob(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def run():
        pass

    @abstractmethod
    def create_config(model: Model, dataset: Dataset):
        pass


class GCPDeployJob(BaseCloudJob):
    def __init__(self):
        super().__init__()

    def run(self, config):
        run_training_job(config)

    def create_config(self, model: Model, dataset: Dataset, project_id: str, location: str, service_account: str):
        """Create a JSON configuration file from model and dataset attributes."""
        rand_suffix = ''.join(random.choices(
            string.ascii_lowercase + string.digits, k=4))
        bucket_name = f"constellaxion-{project_id}"
        job_config = {
            "model": {
                "model_id": model.id,
                "base_model": model.base_model,
            },
            "dataset": {
                "train": {
                    "local": dataset.train,
                    "cloud": f"{model.id}/data/train.csv"
                },
                "val": {
                    "local": dataset.val,
                    "cloud": f"{model.id}/data/val.csv"
                },
                "test": {
                    "local": dataset.test,
                    "cloud": f"{model.id}/data/test.csv"
                },
            },
            "deploy": {
                "provider": "gcp",
                "project_id": project_id,
                "location": location,
                "bucket_name": bucket_name,
                "staging_dir": f"{model.id}/staging",
                "experiments_dir": f"{model.id}/experiments",
                "model_path": f"{model.id}/model",
                "service_account": service_account
            }
        }
        with open("job.json", "w") as f:
            json.dump(job_config, f, indent=4)


class AWSDeployJob(BaseCloudJob):
    def __init__(self, ):
        super().__init__()
        pass

    def run(self):
        pass

    def create_config(self, model: Model, dataset: Dataset):
        pass



import json
from google.cloud import aiplatform, storage
from constellaxion.services.gcp.model_map import model_map
from google.cloud import storage
import pkg_resources


def upload_data_to_gcp(config: dict):
    """
    Upload dataset to GCP, ensuring the bucket exists.

    Args:
        config (dict): Configuration dictionary with bucket and dataset details.
    """
    client = storage.Client()
    bucket_name = config['deploy']['bucket_name']

    # Check if bucket exists
    bucket = client.bucket(bucket_name)
    if not bucket.exists():
        print(f"Bucket '{bucket_name}' does not exist. Creating it...")
        bucket = client.create_bucket(bucket_name)
        print(f"Bucket '{bucket_name}' created successfully.")

    # Upload training dataset
    train_blob = bucket.blob(config['dataset']['train']['cloud'])
    train_blob.upload_from_filename(config['dataset']['train']['local'])
    print(f"Uploaded training dataset to {train_blob.name}")

    # Upload validation dataset
    val_blob = bucket.blob(config['dataset']['val']['cloud'])
    val_blob.upload_from_filename(config['dataset']['val']['local'])
    print(f"Uploaded validation dataset to {val_blob.name}")

    # Upload test dataset
    test_blob = bucket.blob(config['dataset']['test']['cloud'])
    test_blob.upload_from_filename(config['dataset']['test']['local'])
    print(f"Uploaded test dataset to {test_blob.name}")


def create_custom_job_with_experiment_autologging_sample(
        project: str,
        location: str,
        staging_bucket: str,
        display_name: str,
        script_path: str,
        container_uri: str,
        service_account: str,
        requirements: str,
        machine_type: str,
        accelerator_type: str,
        accelerator_count: int,
        replica_count: int,
        args: list[str]
) -> None:
    aiplatform.init(project=project, location=location,
                    staging_bucket=staging_bucket)
    job = aiplatform.CustomJob.from_local_script(
        display_name=display_name,
        script_path=script_path,
        container_uri=container_uri,
        requirements=requirements,
        machine_type=machine_type,
        accelerator_type=accelerator_type,
        accelerator_count=accelerator_count,
        replica_count=replica_count,
        args=args
    )
    job.run(service_account=service_account)


def run_training_job(config):
    model_name = config['model']['base_model']
    bucket_name = config['deploy']['bucket_name']
    script_path = pkg_resources.resource_filename(
        "constellaxion.models.tinyllama_1b.gcp", "lora.py")
    # script_path = model_map[model_name]["lora"]
    infra_config = model_map[model_name]["infra"]
    upload_data_to_gcp(config)
    create_custom_job_with_experiment_autologging_sample(
        project=config['deploy']['project_id'],
        location=config['deploy']['location'],
        staging_bucket=f"gs://{bucket_name}/{config['deploy']['staging_dir']}",
        display_name=config['model']['model_id'],
        script_path=script_path,
        requirements=infra_config['requirements'],
        container_uri=infra_config['container_uri'],
        service_account=config['deploy']['service_account'],
        machine_type=infra_config['machine_type'],
        accelerator_type=infra_config['accelerator_type'],
        accelerator_count=infra_config['accelerator_count'],
        replica_count=infra_config['replica_count'],
        args=[
            f"--train-set={config['dataset']['train']['cloud']}",
            f"--val-set={config['dataset']['val']['cloud']}",
            f"--test-set={config['dataset']['test']['cloud']}",
            f"--bucket-name={config['deploy']['bucket_name']}",
            f"--model-path={config['deploy']['model_path']}",
            f"--experiments-dir={config['deploy']['experiments_dir']}",
            f"--staging-dir={config['deploy']['staging_dir']}"
        ]
    )



import os
import yaml
import click
from constellaxion.handlers.model import Model
from constellaxion.handlers.dataset import Dataset
from constellaxion.handlers.cloud_job import GCPDeployJob
from constellaxion.services.gcp.iam import create_service_account


def init_model(model_config):
    """Initialize the model

    Args:
        model_config (dict): Model config details
    """
    id = model_config.get('id')
    base = model_config.get('base')
    if not id:
        click.echo(
            f"Error: Missing value, model.id in model.yaml file", err=True)
    if not base:
        click.echo(
            f"Error: Missing value, model.base in model.yaml file", err=True)
    return Model(id, base)


def init_dataset(dataset_config):
    """Initialize the dataset

    Args:
        dataset_config (dict): Dataset config details
    """
    train = dataset_config.get('train')
    val = dataset_config.get('val')
    test = dataset_config.get('test')
    if not train:
        click.echo(
            f"Error: Missing value, dataset.train in model.yaml file", err=True)
    if not val:
        click.echo(
            f"Error: Missing value, dataset.val in model.yaml file", err=True)
    if not test:
        click.echo(
            f"Error: Missing value, dataset.test in model.yaml file", err=True)
    return Dataset(train, val, test)


def init_job(model: Model, dataset: Dataset, job_config):
    """Initialize the deployment job definition

    Args:
        job_config (list): List of dicts containing deployment job config details
    """
    gcp = job_config.get('gcp')
    if gcp:
        project_id = gcp.get('project_id')
        location = gcp.get('location')
        if not project_id:
            click.echo(
                f"Error: Missing value, job.gcp.project_id in model.yaml file", err=True)
        if not location:
            click.echo(
                f"Error: Missing value, job.gcp.location in model.yaml file", err=True)

        click.echo(f"Initializing resources for project: {project_id}")
        try:
            service_account_email = create_service_account(project_id)
            if service_account_email:
                click.echo(
                    "The required GCP Service Account is ready to use 🦾")
        except Exception as e:
            click.echo(f"Error: {str(e)}", err=True)
        job = GCPDeployJob()
        # Create job config
        job.create_config(model, dataset, project_id,
                          location, service_account_email)


@click.command(help="Initialize a new model")
def init():
    """
    Initialize a new model
    """
    model_config = os.path.join(os.getcwd(), "model.yaml")
    if not os.path.exists(model_config):
        click.echo(
            "Error: model.yaml file not found in current directory.", err=True)
        return

    click.echo("Preparing new model job 🤖")
    try:
        with open(model_config, 'r') as file:
            config = yaml.safe_load(file)
            # Get configs
            model_config = config.get('model')
            dataset_config = config.get('dataset')
            deploy_config = config.get('deploy')
            # Init configs
            model = init_model(model_config)
            dataset = init_dataset(dataset_config)
            init_job(model, dataset, deploy_config)
            click.echo(
                click.style("Job Config created. Run 'constellaXion job view' to see details or 'constellaXion job run' to start training your model", fg="green"))
        # Parse values and excecute commands
    except yaml.YAMLError as e:
        click.echo(f"Error parsing model.yaml: {str(e)}", err=True)
    except Exception as e:
        click.echo(f"Unexpected error: {str(e)}", err=True)



from constellaxion.types.model_types import BaseModelName, TrainType


class Model:
    def __init__(self, id: str, base_model: BaseModelName):
        self.id = id
        self.base_model = base_model



import click
from constellaxion.services.gcp.iam import create_service_account


@click.command(help="Sync a new project to Osyris")
@click.option('--project-id', required=True, help='GCP Project ID')
def add(project_id: str):
    """
    Add a project
    """
    click.echo(f"Initializing resources for project: {project_id}")
    try:
        create_service_account(project_id)
        click.echo(
            "constellaxion-admin service account successfully created and roles assigned!")
    except Exception as e:
        click.echo(f"Error: {str(e)}", err=True)



from setuptools import setup, find_packages

setup(
    name='constellaxion',
    version='0.3.1',
    packages=find_packages(),
    install_requires=[
        'requests',
        'rich',
        'click',
        "google-api-python-client",
        "google-auth",
        "google-auth-httplib2",
    ],
    entry_points={
        'console_scripts': [
            'constellaxion=constellaxion.main:cli',
        ],
    },
    author='Constellaxion Technologies, Inc.',
    author_email='dev@constellaxion.ai',
    description='The constellaXion CLI for managing your laboratory database',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    keywords=["constellaxion", "ai", "ml ops"],
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.10',
    package_data={"constellaxion": ["models/tinyllama_1b/gcp/*.py"]}
)
